type arbre = Zero | Noeud of arbre * arbre ;;

(* Question 1 *)
let rec expo2 n =
	match n with
	| 0 -> 1
	| _ -> let x = expo2 (n/2) in
		if n mod 2 == 0 then x*x
		else x*x*2
;;

(* Question 2 *)
let rec evalue a = 
	match a with
	| Zero 		-> 0
	| Noeud(g,d) 	-> (expo2 (evalue g)) + (evalue d)
;;

(* Question 4 *)
exception Invalid_Argument;;

let rec dec2 n = match n with
    | 0 -> raise Invalid_Argument
    | 1 -> (0, 0)
    | _ -> let (p, q) = dec2 (n / 2) in (p + 1, if n mod 2 = 0 then 2 * q else 2 * q + 1);;




(* Question 5 *)
let rec arbre_mini n = 
	match n with
	| 0 		-> Zero
	| k when k > 0 	-> let (p,q) = dec2 n in Noeud(arbre_mini p, arbre_mini q)
	| _ -> raise Invalid_Argument
;;

(* Question 6 *)
let rec fermat n =
	let a = arbre_mini n in
	Noeud(Noeud(a,Zero),Noeud(Zero,Zero))
;;

(* on montre que |mu(a)| <= 2a+1 par induction car |mu(3)| = 7 et |mu(1)| = 2, on en dduit que 
|mu(Fn)| <= 2n+7 alors que son criture binaire a 2^n coefficients*)


(* Question 7 *)
let rec egaux a1 a2 = 
	match (a1,a2) with
	| (Zero,Zero) -> true
	| (Noeud(g1,d1),Noeud(g2,d2)) -> (egaux g1 g2) && (egaux d1 d2)
	| _ -> false
;;

let egaux_int n m = egaux (arbre_mini n) (arbre_mini m);;


(* Question 8 *)
(* Complexit linaire par rapport  la somme des tailles des deux arbres*)

(* Question 9 *)
let rec inferieur a1 a2 = 
	match (a1,a2) with
	| (Zero,_) 		-> true
	| (Noeud(_,_),Zero) 	-> false
	| (Noeud(g1,d1),Noeud(g2,d2)) -> 
		if egaux g1 g2 then inferieur d1 d2
		else inferieur g1 g2
;;

let inferieur_int x1 x2 =
	inferieur (arbre_mini x1) (arbre_mini x2)
;;

inferieur_int 45 63;;
inferieur_int 43 12;;

(* Question 10 *)

let rec addition a1 a2 = match a1 with
    | Zero -> a2
    | Noeud (g, d) -> Noeud (g, addition d a2);;


let addition_int x1 x2 =
	evalue (addition (arbre_mini x1) (arbre_mini x2))
;;



(* Question 11 *)


let rec produit a1 a2 =
	let rec prod2 a b =
		match (a,b) with
		| (_,Zero) -> Zero
		| (Zero,_) -> b
		| (_,Noeud(gb,db)) -> Noeud((addition a gb),prod2 a db)
	in
	let rec aux a b =
		match (a,b) with
		| (Zero,_) -> Zero
		| (_,Zero) -> Zero
		| (Noeud(ga,da),Noeud(gb,db)) -> 
			let x1 = addition ga gb
			and x2 = prod2 ga db
			and x3 = prod2 gb da
			and x4 = aux da db
			in addition (Noeud(x1,x4)) (addition x2 x3)
	in aux a1 a2
;;

let produit_int x1 x2 =
	evalue (produit (arbre_mini x1) (arbre_mini x2))
;;


(* Question 12 *)
let rec ajoute_un a = 
	match a with
	| Zero -> Noeud(Zero,Zero)
	| Noeud(g,d) -> let dp1 = ajoute_un d
			in if egaux (Noeud(g,Zero)) dp1 then 
				Noeud(ajoute_un g,Zero)
			else	Noeud(g,dp1)
;;

let ajoute_un_int x =
	evalue (ajoute_un (arbre_mini x))
;;

ajoute_un_int 45;;
ajoute_un_int 63;;

(* Question 13 *)

let optimise a =
	let rec arbre_to_list a = match  a with
	| Zero 		-> []
	| Noeud(g,d) 	-> insert g (arbre_to_list d)
	and insert a l = match l with
	| [] 		-> [a]
	| b::s		-> if egaux a b then insert (ajoute_un a) s
			   else if inferieur a b then a::l
			   else b::(insert a s)
	in 
	let rec list_to_arbre l = match l with
	| [] 		-> Zero
	| x::y 		-> Noeud(x,list_to_arbre y)
	in 
	list_to_arbre (List.rev(arbre_to_list a))
;;

let addition_opt a b = 
	optimise (addition a b)
;;

let addition_opt_int x y =
	evalue (addition_opt (arbre_mini x) (arbre_mini y))
;;

addition_opt_int 45 63;;
addition_opt_int 43 12;;

let rec est_minimal a = 
	match a with
	| Zero 		-> true
	| Noeud(g,d) 	-> (inferieur d (Noeud(g,Zero)))
			&& (not (egaux d (Noeud(g,Zero))))
			&& (est_minimal g)
			&& (est_minimal d)
;;

est_minimal (optimise (addition (arbre_mini 45) (arbre_mini 63)));;
est_minimal (optimise (addition (arbre_mini 43) (arbre_mini 12)));;


